#include <algorithm>
#include <assert.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include <numeric>
#include "CartesianProductDb.hh"
#include "CdbStatic.hh"


CartesianProductDb::CartesianProductDb()
{
}

CartesianProductDb::~CartesianProductDb()
{
}

CartesianProductDb::iterator CartesianProductDb::begin()
{
  return db.begin();
}

CartesianProductDb::const_iterator CartesianProductDb::begin() const
{
  return db.begin();
}

CartesianProductDb::iterator CartesianProductDb::end()
{
  return db.end();
}
  
CartesianProductDb::const_iterator CartesianProductDb::end() const
{
  return db.end();
}

void CartesianProductDb::push_back(const CartesianProduct& cp)
{
  db.push_back(cp);
}

void CartesianProductDb::push_back_num_cov(uint64_t cells_cov)
{
  tot_cells_cov.push_back(cells_cov);
}

void CartesianProductDb::clear()
{
  db.clear();
}

uint64_t CartesianProductDb::size_mem() const
{
  uint64_t s = 0;
  for (backing_t::const_iterator cIt = db.begin(); cIt != db.end(); ++cIt)
  {
    s += cIt->size();
  }
  return s;
}

uint64_t CartesianProductDb::size_k() const
{
  return db.size();
}

std::vector<int> *CartesianProductDb::item_frequency() const
{
  std::vector<int> *i_ct = new std::vector<int>;
  for (backing_t::const_iterator cpIt=db.begin(); cpIt!=db.end(); ++cpIt)
  {
    for (Itemset::const_iterator iIt=cpIt->itemset.begin(); iIt!=cpIt->itemset.end(); ++iIt)
    {
      const int item = *iIt;
      if ((int)i_ct->size() <= item)
      {
        i_ct->resize(item + 1, 0);
      }
      ++i_ct->operator[](item);
    }
  }
  return i_ct;
}

std::vector<int> *CartesianProductDb::transaction_frequency() const
{
  std::vector<int>* t_ct = new std::vector<int>;
  for (backing_t::const_iterator cpIt=db.begin(); cpIt!=db.end(); ++cpIt)
  {
    for (Transactionset::const_iterator tIt=cpIt->transactionset.begin(); tIt!=cpIt->transactionset.end(); ++tIt)
    {
      const int transaction = *tIt;
      if ((int)t_ct->size() <= transaction)
      {
        t_ct->resize(transaction + 1, 0);
      }
      ++t_ct->operator[](transaction);
    }
  }
  return t_ct;
}

void CartesianProductDb::write_map_to_freq_file(const std::vector<int>& cv, const char *f_name) const
{
  std::vector<std::pair<int,int> > vp_map(cv.size());
  for (std::vector<int>::size_type i=0; i<cv.size(); ++i)
  {
    vp_map[i] = std::make_pair(cv[i], i);
  }
  std::sort(vp_map.begin(), vp_map.end());

  std::fstream is_out(f_name, std::ios::out | std::ios::binary);
  
  for (std::vector<std::pair<int,int> >::reverse_iterator vpIt=vp_map.rbegin(); vpIt!=vp_map.rend(); ++vpIt)
  {
    is_out << vpIt->second << "\t" << vpIt->first << "\n";
  }

  is_out.close();
}

void CartesianProductDb::write_frequency_files() const
{
  std::vector<int> *i_ct = item_frequency();
  std::vector<int> *t_ct = transaction_frequency();

  write_map_to_freq_file(*i_ct, "frequency_items");
  write_map_to_freq_file(*t_ct, "frequency_transactions");

  delete i_ct;
  delete t_ct;
}

double CartesianProductDb::set_mdl(const std::vector<int>* freq_list) const
{
  /*
  std::cout << " computing mdl of ";
  std::copy(freq_list->begin(), freq_list->end(), std::ostream_iterator<int>(std::cout,", "));
  std::cout << "\n";
  */

  double total_freq = std::accumulate(freq_list->begin(), freq_list->end(), 0);

  double mdl_sum = 0;
  for (std::vector<int>::const_iterator fIt=freq_list->begin(); fIt != freq_list->end(); ++fIt)
  {
    double cur_freq = *fIt;
    if (cur_freq != 0)
    {
      mdl_sum += cur_freq * log2(cur_freq / total_freq);
    }
  }
  mdl_sum *= -1;

  return mdl_sum;
}

CartesianProductDb::size_type CartesianProductDb::size() const
{
  return db.size();
}

// initially, uncov_zero_db contains set bits for only zeroes who are uncovered (and thus incur a cost)
std::vector<CartesianProductDb::size_type> CartesianProductDb::best_two_merge_candidates(const BitDb& uncov_zero_db) const
{
  double max_aff = -1.0;
  size_type max_aff_i = 0, max_aff_j = 0;

  const long int num_evaluations = size() * (size() - 1) / 2;

  const long int target_samples = std::min(num_evaluations, (long int)(2000 + size()));

  const double eval_prob = (size() < sqrt(target_samples)) ? 1.0 : (double)target_samples / (double)num_evaluations;

  long int num_sampled = 0;

  for (std::vector<float>::size_type i=0; i<size(); ++i)
  {
    for (std::vector<float>::size_type j=0; j < i; ++j)
    {
      if (drand48() < eval_prob)
      {
        ++num_sampled;

        const CartesianProduct& item_i = db[i];
  
        const CartesianProduct& item_j = db[j];
  
        const CartesianProduct ij_union = item_i.union_with(item_j);
  
        int64_t savings = item_j.transactionset.size() + item_i.transactionset.size() + item_i.itemset.size() + item_j.itemset.size() - ij_union.transactionset.size() - ij_union.itemset.size();
  
        uint64_t false_positives = 0;
        for (Itemset::const_iterator iuIt = ij_union.itemset.begin(); iuIt != ij_union.itemset.end(); ++iuIt)
        {
          for (Transactionset::const_iterator tuIt = ij_union.transactionset.begin(); tuIt != ij_union.transactionset.end(); ++ tuIt)
          {
            if (uncov_zero_db.exists(*tuIt, *iuIt))
            {
              ++false_positives;
            }
          }
        }
  
        const double affinity = (double)savings / (double)false_positives;
  
        if (affinity > max_aff)
        {
          max_aff = affinity;
          max_aff_i = i;
          max_aff_j = j;
        }
      }
    }
  }

  //std::cout << "sampled " << num_sampled << " of " << num_evaluations << "\n";

  assert(num_sampled >= target_samples / 2);

  //std::cout << "Merge candidates with max affinity (" << max_aff << ") are " << max_aff_i << " and " << max_aff_j << "\n";

  std::vector<size_type> m_cand;
  m_cand.push_back(max_aff_i);
  m_cand.push_back(max_aff_j);
  return m_cand;
}

std::vector<CartesianProductDb::size_type> CartesianProductDb::best_three_merge_candidates(const BitDb& uncov_zero_db) const
{
  double max_aff = -1.0;
  size_type max_aff_i = 0, max_aff_j, max_aff_k = 0;

  for (std::vector<float>::size_type i=0; i<size(); ++i)
  {
    for (std::vector<float>::size_type j=0; j < i; ++j)
    {
      for (std::vector<float>::size_type k=0; k < j; ++k)
      {
        const CartesianProduct& item_i = db[i];
  
        const CartesianProduct& item_j = db[j];

	const CartesianProduct& item_k = db[k];
  
        const CartesianProduct ijk_union = item_i.union_with(item_j).union_with(item_k);
  
        int64_t savings = item_i.transactionset.size() + item_j.transactionset.size() + item_k.transactionset.size() + item_i.itemset.size() + item_j.itemset.size() + item_k.itemset.size() - ijk_union.transactionset.size() - ijk_union.itemset.size();
  
        uint64_t false_positives = 0;
        for (Itemset::const_iterator iuIt = ijk_union.itemset.begin(); iuIt != ijk_union.itemset.end(); ++iuIt)
        {
          for (Transactionset::const_iterator tuIt = ijk_union.transactionset.begin(); tuIt != ijk_union.transactionset.end(); ++ tuIt)
          {
            if (uncov_zero_db.exists(*tuIt, *iuIt))
            {
              ++false_positives;
            }
          }
        }
  
        const double affinity = (double)savings / (double)false_positives;
  
        if (affinity > max_aff)
        {
          max_aff = affinity;
          max_aff_i = i;
          max_aff_j = j;
	  max_aff_k = k;
        }
      }
    }
  }

  std::vector<size_type> m_cand;
  m_cand.push_back(max_aff_i);
  m_cand.push_back(max_aff_j);
  m_cand.push_back(max_aff_k);
  return m_cand;
}

void CartesianProductDb::merge(BitDb& uncov_zero_db, uint64_t min_cartes_prod, uint64_t max_false_pos, int candidates)
{
  assert(candidates == 2 || candidates == 3);
  //std::cout << "Merging cdb until k <= " << min_cartes_prod << " or num_false_pos >= " << max_false_pos << "\n";
  uint64_t num_false_pos = 0;
  std::vector<double> thresholds;
  thresholds.push_back(-1.0); thresholds.push_back(0.1); thresholds.push_back(0.2); thresholds.push_back(0.4);
  double prev_beta = -1.0;
  while (size() > min_cartes_prod && num_false_pos < max_false_pos)
  {
    //std::cout << "performing merge on cdb of size " << size() << ", size_mem " << size_mem() << " with " << num_false_pos << " false positives so far\n";
    //std::cout << *this << "\n";

    // get pair of best merge candidates
    std::vector<size_type> m_cand;
    if (2 == candidates)
    {
      m_cand = best_two_merge_candidates(uncov_zero_db);
    } else if (3 == candidates) {
      m_cand = best_three_merge_candidates(uncov_zero_db);
    }

    //std::cout << "selected merge candidates " << m_cand[0] << " and " << m_cand[1] << "\n";

    // union them into a single cartesian product
    CartesianProduct merged_cp = db[m_cand.at(0)];
    for (std::vector<size_type>::size_type i=1; i<m_cand.size(); ++i)
    {
      merged_cp = merged_cp.union_with(db[m_cand.at(i)]);
    }

    // mark any newly covered zeroes
    for (Itemset::const_iterator iIt = merged_cp.itemset.begin(); iIt != merged_cp.itemset.end(); ++iIt)
    {
      for (Transactionset::const_iterator tIt = merged_cp.transactionset.begin(); tIt != merged_cp.transactionset.end(); ++tIt)
      {
	if (uncov_zero_db.exists(*tIt, *iIt))
	{
	  ++num_false_pos;
          uncov_zero_db.set_zero(*tIt, *iIt);
	}
      }
    }

    int next_size_mem = size_mem();
    for (size_type i=0; i<m_cand.size(); ++i)
    {
      next_size_mem -= db[m_cand[i]].size();
    }
    next_size_mem += merged_cp.size();

    const double next_beta = (double)num_false_pos / (double)next_size_mem;

    for (std::vector<double>::iterator ti = thresholds.begin(); ti != thresholds.end(); ++ti)
    {
      if (prev_beta <= *ti && next_beta >= *ti)
      {
        std::cout << "merge beta " << *ti << ", size_mem " << next_size_mem << ", size_k " << size() - 1 << ", num_false_pos " << num_false_pos << ", elapsed_time_sec " << CdbStatic::elapsed_time_sec() << "\n";
      }
    }

    if (num_false_pos > max_false_pos)
    {
      //std::cout << "num_false_pos " << num_false_pos << " exceeds max_false_pos " << max_false_pos << "; breaking\n";
      break;
    }

    // erase the candidates from the cdb
    std::sort(m_cand.begin(), m_cand.end());
    for (int i=m_cand.size() - 1; i>=0; --i)
    {
      backing_t::iterator it = db.begin();
      std::advance(it, m_cand.at(i));
      db.erase(it);
    }

    // add the merged cartesian product to the cdb
    //std::cout << "adding merged cartesian product:\n" << merged_cp << "\n";
    db.push_back(merged_cp);

    assert((int)size_mem() == next_size_mem);
    
    std::cout << "merge size_k " << size() << ", size_mem " << size_mem() << ", num_false_pos " << num_false_pos << ", beta " << next_beta << "\n";

    prev_beta = next_beta;
  }

  //std::cout << "Merged cdb has k = " << size() << ", num_false_pos = " << num_false_pos << ", mem_size " << size_mem() << "\n";
  //std::cout << *this << "\n";
}

double CartesianProductDb::item_mdl() const
{
  std::vector<int> *i_ct = item_frequency();
  double i_mdl = set_mdl(i_ct);
  delete i_ct;
  return i_mdl;
}

double CartesianProductDb::transaction_mdl() const
{
  std::vector<int> *t_ct = transaction_frequency();
  double t_mdl = set_mdl(t_ct);
  delete t_ct;
  return t_mdl;
}

std::ostream& operator<<(std::ostream& os, const CartesianProductDb& cp_db)
{
  for (CartesianProductDb::backing_t::size_type i=0; i<cp_db.db.size(); ++i)
  {
    //os << i << ": " << cp_db.db[i] << "\n";
    os << cp_db.db[i] << "\n";
  }
  return os;
}
