#include <algorithm>
#include <assert.h>
#include <errno.h>
#include <ext/algorithm>
#include <ext/functional>
#include <fstream>
#include <iostream>
#include <iterator>
#include <limits>
#include <math.h>
#include <queue>
#include <string>
#include <sstream>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include "PTNode.hh"
#include "TreeWriter.hh"
#include "BitDb.hh"
#include "CartesianProductDb.hh"
#include "CdbStatic.hh"

void populate_with_transactions(PTNode *root)
{
  std::queue<PTNode*> nqueue;
  nqueue.push(root);

  while (!nqueue.empty())
  {
    PTNode *node = nqueue.front();
    nqueue.pop();

    for (PTNode::PTChildMap::iterator cIt=node->children.begin(); cIt!=node->children.end(); ++cIt)
    {
      PTNode *child = cIt->second;

      std::vector<Itemset::item_t> inters_vec(child->itemset.size() - 1);
      Itemset::iterator iIt = child->itemset.begin();
      iIt++;
      std::copy(iIt, child->itemset.end(), inters_vec.begin());

      PTNode *inters_node = root->find(inters_vec);
      
      std::back_insert_iterator<Transactionset > ct_ii(child->transactionset);
      
      std::set_intersection(node->transactionset.begin(), node->transactionset.end(), inters_node->transactionset.begin(), inters_node->transactionset.end(), ct_ii);
  
      nqueue.push(child);
    }
  }
}

CartesianProductDb* collapse_cdb_by(const CartesianProductDb* cdb, CartesianProduct::Component component)
{
  PTNode* froot = new PTNode();

  for (CartesianProductDb::const_iterator cdbIt=cdb->begin(); cdbIt!=cdb->end(); ++cdbIt)
  {
    std::set<Itemset::item_t> path;
    if (CartesianProduct::ITEMSET == component)
    {
      path.insert(cdbIt->itemset.begin(), cdbIt->itemset.end());
    } else if (CartesianProduct::TRANSACTIONSET == component) {
      path.insert(cdbIt->transactionset.begin(), cdbIt->transactionset.end());
    }

    PTNode *built_node = froot->build_node(path);

    built_node->itemset.resize(cdbIt->itemset.size());
    std::copy(cdbIt->itemset.begin(), cdbIt->itemset.end(), built_node->itemset.begin());

    built_node->transactionset.resize(cdbIt->transactionset.size());
    std::copy(cdbIt->transactionset.begin(), cdbIt->transactionset.end(), built_node->transactionset.begin());
  }

  CartesianProductDb* nl = froot->to_nodelist();

  delete froot;

  return nl;
}

int main(int argc, char **argv)
{
  if (argc < 3)
  {
    std::cerr << "Usage: " << argv[0] << " dataset support beta [merge_k] [merge_way(2|3)]\n";
    exit(1);
  }

  struct timeval tv;
  gettimeofday(&tv, 0);
  srand48(tv.tv_sec + tv.tv_usec);

  CdbStatic::init_timer();

  errno = 0;
  double support = strtod(argv[2], NULL);
  if (errno != 0)
  {
    std::cerr << "Error converting support " << argv[2] << " to double.\n";
    exit(1);
  }

  if (argc >= 4)
  {
    CdbStatic::beta = strtod(argv[3], NULL);
  }

  if (argc >= 5)
  {
    errno = 0;
    CdbStatic::merge_k = strtol(argv[4], NULL, 10);
    if (0 != errno)
    {
      perror("strtol argv[4]");
      exit(1);
    }
  }
  
  int merge_candidates = 2;
  if (argc >= 6)
  {
    merge_candidates = strtol(argv[5], NULL, 10);
    assert(merge_candidates == 2 || merge_candidates == 3);
  }

  std::cout << "Starting cdb, version 1.0\n";

  CdbStatic::ds_fpath = argv[1];
  std::string::size_type last_slash_pos = CdbStatic::ds_fpath.rfind('/');
  std::string ds_dir(CdbStatic::ds_fpath.substr(0, last_slash_pos + 1));
  std::string ds_fname(CdbStatic::ds_fpath.substr(last_slash_pos + 1));
  std::string ds_fbase(ds_fname.substr(0, ds_fname.find('.')));

  std::string fi_dir(ds_dir + "freq_itemsets");
  if (mkdir(fi_dir.c_str(), 0775) != 0)
  {
    if (errno != EEXIST)
    {
      perror("mkdir freq_itemsets");
      exit(1);
    }
  }

  std::ostringstream fi_fname_str;
  fi_fname_str << fi_dir << "/" << ds_fbase << "_" << support << ".freq";
  const std::string fi_fname(fi_fname_str.str());

  // regenerate frequent itemset file if necessary
  struct stat ds_stat, fi_stat;
  if (stat(CdbStatic::ds_fpath.c_str(), &ds_stat) == -1)
  {
    perror("stat dataset file");
    exit(1);
  }
  if (stat(fi_fname.c_str(), &fi_stat) != -1)
  {
    if (fi_stat.st_mtime <= ds_stat.st_mtime)
    {
      std::cout << "frequent itemset file " << fi_fname << " is not newer than dataset file " << CdbStatic::ds_fpath << "; deleting frequent itemset file\n";
      if (unlink(fi_fname.c_str()) == -1)
      {
        perror("unlink fi_fname");
	exit(1);
      }
    }
  }

  bool no_apriori_res = false;
  FILE *fi_file;
  if ((fi_file = fopen(fi_fname.c_str(), "rb")) == NULL)
  {
    std::cout << "freq itemset file " << fi_fname << " does not exist; creating.\n";
    
    std::ostringstream mk_fi_cmd;
    mk_fi_cmd << "bin/apriori -ts";
    mk_fi_cmd << " -a";
    mk_fi_cmd << " -s" << support;
    mk_fi_cmd << " " << CdbStatic::ds_fpath << " " << fi_fname;
    std::cout << "Executing " << mk_fi_cmd.str() << "\n";
    int fi_cmd_st;
    if ((fi_cmd_st = system(mk_fi_cmd.str().c_str())) == -1)
    {
      std::cerr << "call to `" << mk_fi_cmd.str() << "` failed\n";
      perror("make fi file");
      exit(1);
    }

    no_apriori_res = (WEXITSTATUS(fi_cmd_st) == 241);

    if (no_apriori_res)
    {
      std::cout << "Warning: apriori returns no frequent itemsets with support " << support << "\n";
    }
    if (WEXITSTATUS(fi_cmd_st) != 0 && !no_apriori_res)
    {
       std::cerr << "command `" << mk_fi_cmd.str() << "` returned " << WEXITSTATUS(fi_cmd_st) << "\n";
       exit(1);
    }

    if (!no_apriori_res)
    {
      if ((fi_file = fopen(fi_fname.c_str(), "rb")) == NULL)
      {
        perror("fopen fi_fname");
        exit(1);
      }
    }
  }

 std::cout << "using freq itemset file " << fi_fname << "\n";

  size_t line_size = 4096;
  PTNode* root = new PTNode();

  if (!no_apriori_res)
  {
    char *fi_line;
    if ((fi_line = (char*)malloc(line_size)) == NULL)
    {
      perror("malloc fi_line");
      exit(1);
    }
  
    while (getline(&fi_line, &line_size, fi_file) != -1)
    {
      //std::cout << "Reading line " << fi_line;
      int item_digit_len, item_int;
      std::set<Itemset::item_t> items;
      char *fi_line_scan = fi_line;
      while (sscanf(fi_line_scan, "%d%n", &item_int, &item_digit_len) == 1)
      {
        if (item_int < std::numeric_limits<Itemset::item_t>::min() || item_int > std::numeric_limits<Itemset::item_t>::max())
	{
          std::cout << "Item out of range: " << item_int << "\n";
	  exit(1);
	}
	const Itemset::item_t item = static_cast<Itemset::item_t>(item_int);
        items.insert(item);
        fi_line_scan += item_digit_len;
      }
      float rel_support;
      int abs_support;
      if (sscanf(fi_line_scan, "  (%f%%/%d)", &rel_support, &abs_support) != 2)
      {
        std::cerr << "Error parsing relative & absolute supports from '" << fi_line_scan << "'\n";
        std::cerr << "Expected a string like \"  (rel_support%/abs_support)\"\n";
        exit(1);
      }
  
      PTNode *ptNode = root->build_node(items);
      ptNode->itemset.resize(items.size());
      std::copy(items.begin(), items.end(), ptNode->itemset.begin());

      ptNode->gamma = (float)(items.size() + abs_support) / (float)(items.size() * abs_support);
      //std::cout << "adding node " << ptNode << " with calculated gamma " << ptNode->gamma << " from items.size() " << items.size() << ", abs_support " << abs_support << "\n";
    }
  
    free(fi_line);
  
    if (fclose(fi_file) == -1)
    {
      perror("fclose fi_file");
      exit(1);
    }
  }


  FILE *ds_file;
  if ((ds_file = fopen(CdbStatic::ds_fpath.c_str(), "rb")) == NULL)
  {
    std::cerr << CdbStatic::ds_fpath << "\n";
    perror("fopen ds_fpath");
    exit(1);
  }

  char *ds_line;
  if ((ds_line = (char*)malloc(line_size)) == NULL)
  {
    perror("malloc ds_line");
    exit(1);
  }

  // populate first level of nodes with transactions by scanning the dataset
  CdbStatic::max_item = std::numeric_limits<Itemset::item_t>::min();
  CdbStatic::min_item = std::numeric_limits<Itemset::item_t>::max();

  uint64_t trans_num_int64 = 0, tot_cells_num = 0;
  int item_int, item_digit_len;
  char *ds_line_scan;
  PTNode::PTChildMap::iterator item_child;
  while (getline(&ds_line, &line_size, ds_file) != -1)
  {
    ++trans_num_int64;
    if (trans_num_int64 < std::numeric_limits<Transactionset::transaction_t>::min() || trans_num_int64 > std::numeric_limits<Transactionset::transaction_t>::max())
    {
      std::cout << "Transaction id out of range: " << trans_num_int64 << "\n";
      exit(1);
    }
    ds_line_scan = ds_line;
    while (sscanf(ds_line_scan, "%d%n", &item_int, &item_digit_len) != EOF)
    {
      if (item_int < std::numeric_limits<Itemset::item_t>::min() || item_int > std::numeric_limits<Itemset::item_t>::max())
      {
        std::cout << "Error: item out of range: " << item_int << "\n";
	exit(1);
      }
      const Itemset::item_t item = static_cast<Itemset::item_t>(item_int);
      //std::cout << "Processing transaction " << trans_num << ", item " << item << ", item_digit_len " << item_digit_len << "\n";
      CdbStatic::max_item = std::max(CdbStatic::max_item, item);
      CdbStatic::min_item = std::min(CdbStatic::min_item, item);

      PTNode *child;
      if ((item_child = root->children.find(item)) == root->children.end())
      {
	std::set<Itemset::item_t> single_item;
	single_item.insert(item);
	child = root->build_node(single_item);
	child->itemset.push_back(item);
	//std::cout << "built single node " << child << "\n";
      } else {
        child = item_child->second;
	//std::cout << "retreived single node " << child << "\n";
      }
      //std::cout << "  child " << item << " matches\n";
      child->transactionset.push_back(static_cast<Transactionset::transaction_t>(trans_num_int64));
      ds_line_scan += item_digit_len;
      ++tot_cells_num;
    }
  }

  CdbStatic::num_transactions = static_cast<Transactionset::transaction_t>(trans_num_int64);
  
  // populate with a bit for every zero
  BitDb uncov_zero_db(CdbStatic::num_transactions, CdbStatic::min_item, CdbStatic::max_item);

  if (fseek(ds_file, 0, SEEK_SET) != 0)
  {
    perror("fseek ds_file");
    exit(1);
  }

  Transactionset::transaction_t cur_trans = 0;
  while (getline(&ds_line, &line_size, ds_file) != -1)
  {
    ++cur_trans;
    std::vector<bool> line_items(CdbStatic::max_item);
    ds_line_scan = ds_line;
    while (sscanf(ds_line_scan, "%d%n", &item_int, &item_digit_len) != EOF)
    {
      const Itemset::item_t item = static_cast<Itemset::item_t>(item_int);
      line_items[item] = true;
      ds_line_scan += item_digit_len;
    }
    for (Itemset::item_t i=CdbStatic::min_item; i<=CdbStatic::max_item; ++i)
    {
      if (!line_items[i])
      {
        uncov_zero_db.insert(cur_trans, i);
      }
    }
  }

  free(ds_line);

  if (fclose(ds_file) == -1)
  {
    perror("fclose ds_fpath");
    exit(1);
  }

  //std::cout << "uncov_zero_db:\n" <<  uncov_zero_db << "\n";

  for (PTNode::PTChildMap::iterator rcIt=root->children.begin(); rcIt!=root->children.end(); ++rcIt)
  {
    PTNode *child = rcIt->second;
    //std::cout << "checking root child " << child << "\n";
    if (child->gamma == std::numeric_limits<float>::max())
    {
      child->gamma = (float)(child->transactionset.size() + child->itemset.size()) / (float)(child->transactionset.size() * child->itemset.size());
      //std::cout << "updated root child's init_support gamma to " << child->gamma << "\n";
    }
  }

  std::cout << "orig db size_horiz: " << tot_cells_num + CdbStatic::num_transactions << "\n";
  std::cout << "orig db size_min: " << CdbStatic::num_transactions + (CdbStatic::max_item - CdbStatic::min_item + 1) << "\n";
  //std::cout << "orig db " << ds_fname << " has " << trans_num << " transactions and " << tot_items_num << " total items\n";
  //std::cout << "orig size: " << tot_items_num * 2 << " (<TID,ITEM_ID> metric)\n";
  //std::cout << "orig size: " << trans_num + tot_items_num << " (<TID, ITEM_ID_1, ITEM_ID_2, ... > metric)\n";

  //populate_with_transactions(root);

  BitDb *covered = new BitDb(CdbStatic::num_transactions, CdbStatic::min_item, CdbStatic::max_item);

  typedef PTNode::GONodeSet GONodeSet;

  GONodeSet* gons = root->gamma_ordered_nodeset();

  CartesianProductDb cdb;

  /*
  char cdb_outp_n[1024];
  strcpy(cdb_outp_n, CdbStatic::ds_fpath.c_str());
  cdb_outp_n[CdbStatic::ds_fpath.length() - 4] = '\0';
  char support_str[1024];
  strcat(cdb_outp_n, "_");
  sprintf(support_str, "%0.1f", support);
  strcat(cdb_outp_n, support_str);
  strcat(cdb_outp_n, ".hyper");

  std::cout << "writing data to " << cdb_outp_n << "\n";

  std::ofstream cdb_outp(cdb_outp_n, std::ios::out | std::ios::trunc | std::ios::binary);
  if (!cdb_outp.good())
  {
    perror("ofstream cdb_outp");
    exit(1);
  }
  //int cdb_outp_lines = 0;
  */

  while(!gons->empty())
  {
    /*
    std::cout << "\n";
    std::cout << "gamma-ordered queue:\n";
    std::copy(gons->begin(), gons->end(), std::ostream_iterator<PTNode*>(std::cout,"\n"));
    */

    // pop first element from gamma-ordered queue
    PTNode *node;
    {
      GONodeSet::iterator g_front = gons->begin();
      node = *g_front;
      gons->erase(g_front);
    }

    // recalculate gamma for node & store chosen transactions
    Transactionset* trans_chosen = new Transactionset;

    node->recalc_gamma(*covered, trans_chosen, root); 

    /*
    std::cout << "before entering loop, using top node " << node << "\n";
    */

    while(!gons->empty() && (node->gamma > (*gons->begin())->gamma || node->gamma == std::numeric_limits<float>::max()))
    {
      //std::cout << "considering node " << node << "\n";

      if (node->gamma != std::numeric_limits<float>::max())
      {
        gons->insert(node);
      }

      node = *gons->begin();

      size_t gons_prev_size = gons->size();
      gons->erase(gons->begin());
      assert(gons->size() == gons_prev_size - 1);

      //std::cout << "new node " << node;

      node->recalc_gamma(*covered, trans_chosen, root);
      
      //std::cout << " gamma recalculated to " << node->gamma << "\n";
    }

    if (gons->empty() && node->gamma == std::numeric_limits<float>::max())
    {
      delete trans_chosen;
      trans_chosen = NULL;
      //std::cout << "No useful transactions remain; breaking.\n";

      break;
    }

    //assert(__gnu_cxx::is_sorted(trans_chosen->begin(), trans_chosen->end()));

    /*
    std::cout << "chose node " << node << "\n";

    std::cout << "adding to cdb: { ";
    for (Itemset::const_iterator iIt=node->itemset.begin(); iIt!=node->itemset.end(); ++iIt)
    {
      if (iIt != node->itemset.begin()) std::cout << ", ";
      std::cout << *iIt;
    }
    std::cout << " }, [ ";
    for (Transactionset::const_iterator tIt=trans_chosen->begin(); tIt!=trans_chosen->end(); ++tIt)
    {
      if (tIt != trans_chosen->begin()) std::cout << ", ";
      std::cout << *tIt;
    }
    std::cout << " ]\n";
    */
    
    CartesianProduct cp(node->itemset, *trans_chosen);
    cdb.push_back(cp);

    /*
    ++cdb_outp_lines;
    //std::cout << "outputting line " << cdb_outp_lines << ": " << cp << "\n";
    cdb_outp << cp << "\n";

    if (cdb_outp.bad())
    {
      perror("cdb_outp <<");
      exit(1);
    }

    cdb_outp.flush();

    if (cdb_outp_lines == 20)
    {
      std::cout << "cdb output complete; exiting\n";
      exit(0);
    }
    */
    
    Transactionset::size_type num_trans_covered = 0;
    for (Transactionset::const_iterator tIt=node->transactionset.begin(); tIt != node->transactionset.end(); ++tIt)
    {
      Transactionset::transaction_t transaction = *tIt;

      /*
      {
        Transactionset::iterator ntIt = std::find(node->transactionset.begin(), node->transactionset.end(), transaction);
        assert(ntIt != node->transactionset.end());
        node->transactionset.erase(ntIt);
      }
      */

      if (std::find(trans_chosen->begin(), trans_chosen->end(), transaction) != trans_chosen->end())
      {
        ++num_trans_covered;
        for (Itemset::const_iterator iIt=node->itemset.begin(); iIt != node->itemset.end(); ++iIt)
        {
          Itemset::item_t item = *iIt;
          //std::cout << "marking { " << transaction << ", " << item << " } as covered; ";
          covered->insert(transaction, item);
          //std::cout << "covered.size() is now " << covered.size() << "\n";
        }
      } else {
        bool all_items_chosen = true;
        for (Itemset::const_iterator iIt=node->itemset.begin(); iIt != node->itemset.end(); ++iIt)
        {
          Itemset::item_t item = *iIt;
	  if (!covered->exists(transaction, item))
	  {
            all_items_chosen = false;
	    break;
	  }
	}
	if (all_items_chosen)
	{
          ++num_trans_covered;
	}
      }
    }

    assert(num_trans_covered <= node->transactionset.size());

    //cdb.push_back_num_cov(covered->num_bits_set());
    
    //std::cout << "after adding cp " << CartesianProduct(node->itemset, *trans_chosen) << ", num_bits_set is " << covered->num_bits_set() << "\n";

    //std::cout << "coverage matrix is:\n" << covered << "\n";

    delete trans_chosen;
    trans_chosen = NULL;

    if (node->transactionset.size() > num_trans_covered)
    {
      node->recalc_gamma(*covered, NULL);
      //std::cout << "recalculated " << node << "\n";
      gons->insert(node);
    }
  }
  
  //cdb_outp.close();

  delete covered;

  delete gons;

  const uint64_t cdb_size_mem = cdb.size_mem();
  const uint64_t cdb_size_k = cdb.size();
  std::cout << "cdb size_mem: " << cdb.size_mem() << "\n";
  std::cout << "cdb size_k: " << cdb.size() << "\n";

  double item_bits = ceil(log2(CdbStatic::max_item));
  //std::cout << "need " << item_bits << " bits to represent an item_id\n";
  double item_mdl = cdb.item_mdl() / item_bits;

  double tid_bits = ceil(log2(CdbStatic::num_transactions));
  //std::cout << "need " << tid_bits << " bits to represent a transaction_id\n";
  double transaction_mdl = cdb.transaction_mdl() / tid_bits;

  //std::cout << "cdb mdl: " << item_mdl << " + " << transaction_mdl << " == " << item_mdl + transaction_mdl << "\n";
  std::cout << "cdb mdl: " << item_mdl + transaction_mdl << "\n";
  std::cout << "cdb coverage vector: ";
  for (std::vector<uint64_t>::const_iterator tcIt=cdb.tot_cells_cov.begin(); tcIt != cdb.tot_cells_cov.end(); ++tcIt)
  {
    if (tcIt != cdb.tot_cells_cov.begin()) std::cout << " ";
    std::cout << *tcIt;
  }
  std::cout << "\n";
  /*
  std::cout << "cartesian product list:\n";
  std::copy(cdb.begin(), cdb.end(), std::ostream_iterator<CartesianProduct>(std::cout, "\n"));
  */

  CartesianProductDb* cdb1 = collapse_cdb_by(&cdb, CartesianProduct::ITEMSET);
  //std::cout << "cdb1 size: " << cdb1->size_mem() << " (collapsed by itemset)\n";
  /*
  std::cout << "collapsed by itemset:\n";
  std::copy(cdb1->begin(), cdb1->end(), std::ostream_iterator<CartesianProduct>(std::cout, "\n"));
  */

  cdb.clear();

  CartesianProductDb* cdb2 = collapse_cdb_by(cdb1, CartesianProduct::TRANSACTIONSET);

  delete cdb1;

  std::cout << "collapsed cdb size_mem: " << cdb2->size_mem() << "\n";
  std::cout << "collapsed cdb size_k: " << cdb2->size() << "\n";

  //std::cout << "cdb2 size: " << cdb2->size_mem() << " (collapsed)\n";
  std::cout << "collapse size_mem reduction: " << (double)cdb2->size_mem() / (double)cdb_size_mem << "\n";
  std::cout << "collapse size_k reduction: " << (double)cdb2->size() / (double)cdb_size_k << "\n";
  std::cout << "collapse elapsed_time_sec: " << CdbStatic::elapsed_time_sec() << "\n";

  //std::cout << "pre-merge cdb:\n" << *cdb2 << "\n";

  const uint64_t pre_merge_size_mem = cdb2->size_mem();
  const uint64_t pre_merge_size_k = cdb2->size();

  if (CdbStatic::beta >= 0 || CdbStatic::merge_k < cdb2->size())
  {
    if (CdbStatic::beta >= 0)
    {
      //assert(CdbStatic::merge_k == std::numeric_limits<unsigned int>::max());
      CdbStatic::merge_k = 1;
    } else {
      assert(CdbStatic::beta < 0.0);
    }
    uint64_t false_pos_budget = (uint64_t)round((double)tot_cells_num * CdbStatic::beta);

    cdb2->merge(uncov_zero_db, CdbStatic::merge_k, false_pos_budget, merge_candidates);

    std::cout << "merged cdb size_mem: " << cdb2->size_mem() << "\n";
    std::cout << "merged cdb size_k: " << cdb2->size() << "\n";
    std::cout << "merge size_mem reduction: " << (double)cdb2->size_mem() / (double)pre_merge_size_mem << "\n";
    std::cout << "merge size_k reduction: " << (double)cdb2->size() / (double)pre_merge_size_k << "\n";
    std::cout << "merge elapsed_time: " << CdbStatic::elapsed_time_sec() << "\n";

    //std::cout << "merged cdb:\n" << *cdb2 << "\n";
  }

  std::ostringstream out_cdb;
  out_cdb << CdbStatic::ds_fpath.substr(0, CdbStatic::ds_fpath.length() - 4);
  out_cdb << "_support-" << support;
  out_cdb << "_beta-" << CdbStatic::beta;
  out_cdb << "_mergek-" << CdbStatic::merge_k;
  out_cdb << ".cdb";

  std::cout << "Writing hyperrectangle output to " << out_cdb.str() << "\n";
  std::cout << "(Format is: itemset; transactionset)\n";

  std::ofstream cdb_outp(out_cdb.str().c_str());
  cdb_outp << *cdb2;
  cdb_outp.close();

  delete cdb2;

  delete root;

  return 0;
}
